
/*
 * jslice.c
 *
 * This program will take a slice of a given picture and save it to
 * a second file name
 *
 */

/*
 * CHANGES
 * 
 * 18/10/2000 - Added PNG support, Added everything support  via ImageMagick
 * 31/05/2000 - Added JavaScript enhancements and direct embedding
 * 21/05/2000 - Added image-dimentions
 * 25/05/2000 - Fixed up hyperlink detection from slice file
 * 
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <magick/api.h>

/*
 * Include file for users of JPEG library.
 * You will need to have included system headers that define at least
 * the typedefs FILE and size_t before you can include jpeglib.h.
 * (stdio.h is sufficient on ANSI-conforming systems.)
 * You may also wish to include "jerror.h".
 */


#define _VERSION "1.1.0"
#define _MESSAGE "<!-- ----------------------------------------\n\
	Created with the assistance of jslice, v1.1.0 \n\
	written by Paul L Daniels\n\
	pldaniels@pldaniels.com\n\
	http://www.pldaniels.com/jslice\n\
	(C) P.L.Daniels 2000\n\
	------------------------------------------ -->\n\n"
#define _HELP "jslice v1.1.0 - (C) Paul L Daniels 2000\n\n\
	Options Available -\n\
	-i Input image filename [ie -i fp_active.jpg]\n\
	-p Output filename prefix [ie -c fpa]\n\
	[-s Output filename suffix (denotes file type, ie .png .jpg etc, DEFAULT is .jpg)]\n\
	-c Cut/Slice file [ie -c fp.slices]\n\
	-q Quality of JPEG output [ie -q 85] - no longer used, kept for compatibility\n\
	-h OPTIONAL - Header file to prefix to HTML output [ie -h header.html]\n\
	-f OPTIONAL - Footer file to suffix to HTML output [ie -f footer.html]\n\
	--help produce this help.\n\
	\n\
	example of using jslice...\n\
		jslice -i fp_active.jpg -p fpa -c fp.slices -h header.html -f footer.html > index.html\n\n\
	end of help.\n\n"
	
	
#define MAX_HEIGHT 4096
#define MAX_WIDTH 4096
#define MAX_LINK_SIZE 1024
#define MAX_SLICES 1000

#define _CUT 1
#define _UNCUT 0

#define _QUALITY_DEFAULT 85

#define _ERR_CANTREADSCRIPT 103
#define _ERR_INSUFFICENTMEM 200

int image_height;	/* Number of rows in image */
int image_width;		/* Number of columns in image */
int p_height; 	/* width of image after cut */
int p_width;  	/* height of image after cut */
char localpath[]=".";
char _default_suffix[]=".jpg";
/* define a structure that will keep the details of each box slice.
*/

struct _aslice {

	int ax, bx, ay, by;
	char link[MAX_LINK_SIZE+1];
	};

typedef struct _aslice _tslice;
int slicesx[MAX_WIDTH];
int slicesy[MAX_HEIGHT];


/* And of course, create a global variable out of it.
*/

_tslice slices[MAX_SLICES];







/* get_cut_count_x
 * Determines how many slices we have along the X axis
 */
int get_cut_count_x( int maxx ) {

	int count = 0;
	int x;

	for (x = 0; x < maxx; x++) {
		if (slicesx[x] != _UNCUT) count++;
		}

	return count;
	}

/* get_cut_count_y
 * Determines how many slicemarks we have along the Y axis 
 */
int get_cut_count_y( int maxy ) {

	int count = 0;
	int y;

	for (y = 0; y < maxy; y++) {
		if (slicesy[y] != _UNCUT) count++;
		}

	return count;
	}





int count_cuts( int sx, int fx ) {

	int cuts = 0;
	int cx;

	/* for every pixel across the horizontal...*/
	for (cx = sx; cx < fx; cx++){
	
		/* if we've found a cut mark...*/
		if (slicesx[cx] != _UNCUT) {
		
			/* increase our cut count */
			cuts++;

			/* if the next pixel is also cut, but from the same row
			 * then we must skip it, as it's just consecutive cutouts
			 */
			if (slicesx[cx+1] == slicesx[cx]) { 
				cx++;
				}
				
			} /* if slices != UNCUT */
			
		} /* for */
		
	return cuts;
	}





	/* 
	 * dump_slices
	 *
	 * Debugging and information routine
	 */
int dump_slices( void ) {

	int i;
	fprintf(stdout,"\n<!-- ");
	for (i = 0; i < MAX_WIDTH; i++){
		if (slicesx[i] != _UNCUT) fprintf(stdout,"%d, ",i);
	}
	fprintf(stdout," -->\n");

	return 0;
}





int script_cut( Image *image, ImageInfo *image_info, char *script_name, char *foutprefix, char *foutsuffix ){

#define MAX 1023

	FILE *script;
	int count;
	int ax;
	int i;
	int slicewidth;
	int tally;
	int cols;
	int lastx;
	int ffr;
	int slicew, sliceh;
	int real_url = 0;
	int urltest;
	int linecount = 0;
	int status;
	char line[1024];
	char foutname[1024];
	char *link;

	/* ImageMagick vars */
	ExceptionInfo exception;
	Image *peice;
	ImageInfo *peice_info;	
	RectangleInfo peice_box;
	


	/* Initialise the ImageMagick structures */
/*	MagickIncarnate( .... );*/
	GetExceptionInfo(&exception);
	peice_info=CloneImageInfo((ImageInfo *) NULL);
		
	script = fopen(script_name,"r");
	
	/* 
	 * Make sure we don't try reading dud files 
	 */
	if (script == NULL) {
		fprintf(stderr,"Error when trying to open the slice-coordinate file %s\nPlease check to see that the file name is correct, and that it is a text only file\n",script_name);
		exit(_ERR_CANTREADSCRIPT);
		}
		
	count = 0;
	
	/* clean out the slices rulers */
	for (ax = 0; ax < MAX_WIDTH; ax++) {
		slicesx[ax] = _UNCUT;
		slicesy[ax] = _UNCUT;
		}

	/* get all our lines... */
	while (fgets(line, MAX, script) != NULL) {
	
		linecount++;
		
		/*
		 * Digest our line if it's not a comment 
		 */
		if ((line[0] != '#')&&(strlen(line) > 7)){

			/* get the coordinates... */
			slices[count].ax = atoi(strtok(line," "));
			if (slices[count].ax > image_width) {
				fprintf(stderr,"Line %d: X1 value of %d too large, reduced to %d\n",linecount,slices[count].ax,image_width);
				slices[count].ax = image_width;
				}
			if (slices[count].ax < 0) {
				fprintf(stderr,"Line %d: X1 value of %d too small, value now 0.\n",linecount,slices[count].ax);
				slices[count].ax = 0;
				}

		slices[count].ay = atoi(strtok(NULL," "));
			if (slices[count].ay >= image_height) {
				fprintf(stderr,"Line %d: Y1 value of %d too large, reduced to %d\n",linecount,slices[count].ay,image_height-2);
				slices[count].ay = image_height-2;
				}
			if (slices[count].ay < 0) {
				fprintf(stderr,"Line %d: Y1 value of %d too small, value now 0.\n",linecount,slices[count].ay);
				slices[count].ay = 0;
				}

		slices[count].bx = atoi(strtok(NULL," ")); 
			if (slices[count].bx > image_width) {
				fprintf(stderr,"Line %d: X2 value of %d too large, reduced to %d\n",linecount,slices[count].bx,image_width);
				slices[count].bx = image_width;
				}
			if (slices[count].bx < 0) {
				fprintf(stderr,"Line %d: X2 value of %d too small, value now 0.\n",linecount,slices[count].bx);
				slices[count].bx = 0;
				}

		slices[count].by = atoi(strtok(NULL," "));
			if (slices[count].by >= image_height) {
				fprintf(stderr,"Line %d: Y2 value of %d too large, reduced to %d\n",linecount,slices[count].by,image_height-2);
				slices[count].by = image_height-2;
				}
			if (slices[count].by < 0) {
				fprintf(stderr,"Line %d: Y2 value of %d too small, value now 0.\n",linecount,slices[count].by);
				slices[count].by = 0;
				}
		
		/* get our link... */
		link = strtok(NULL,"\n\r");
		
		/*
		 * Test to see that the URL string is potentially valid
		 *
		 */
		if (link != NULL) {
			for (urltest = 0; urltest < strlen(link); urltest++){
				real_url += isalnum(link[urltest]);
				}
			if (urltest > 0) strcpy(slices[count].link, link);
			}
		else slices[count].link[0] = '\0';
		

		/*
		 * Stamp in the slice arrays our marks where we have to slice
		 *
		 * 0.1.3 - We mark in the top-y value in the ax and bx here
		 * so that we can determin if we are counting one or two
		 * columns when we come to counting up the number of cuts
		 * between a given ax -> bx
		 *
		 */
		slicesx[slices[count].ax] = slices[count].ay +_CUT;
		slicesx[slices[count].bx] = slices[count].ay +_CUT;
		slicesy[slices[count].ay] = _CUT;
		slicesy[slices[count].by] = _CUT;

		count++;
		}
	}

	/*
	 * Let the HTML Table start flowing
	 *
	 */
	fprintf(stdout,"<TABLE border=\"0\" cellpadding=\"0\" cellspacing=\"0\"><TR>\n");
		
	/* 
	 * Print out our SLICING SPACERS! 
	 *
	 */
	slicewidth = tally = 1;
	for ( ax = 1; ax <= image_width; ax++ ){
		if (slicesx[ax] != _UNCUT) {
			if (slicesx[ax-1] != slicesx[ax]){
				fprintf(stdout,"<TD><img src=\"spacer.gif\" width=\"%d\" height=\"1\"></td>",slicewidth);
				tally += slicewidth;
				slicewidth=0;
				}
			}
		slicewidth++;
		}

	fprintf(stdout,"\n<!-- total width = %d -->\n",tally);
	fprintf(stdout,"</tr>\n<TR>\n");

	/* print out commentory information about the slices 
	 * to the HTML file
	 */
	dump_slices();

	lastx = -1;
	ffr = 1;
	tally = 0;
	
	/* Now, for every slice, we just cut and save it from the original JPEG
	 */
	for ( i = 0; i < count; i++ ){

		/* create our filename of the output slice */
		sprintf(foutname,"./%s%d%s",foutprefix,i+1,foutsuffix);
		
		/* determine how many columns we have for this slice */
		cols = count_cuts(slices[i].ax, slices[i].bx);

		/* We should also be adding a "rows" count here, but that'll (for now) blow my mind!
		rows = count_y_cuts(slices[i].ay, slices[i].by);
		 */
		 
		/* Catch the scenario where we've moved onto a new row...the blocks wrap back around */
		if (lastx >= slices[i].ax){
			ffr = 1;
			fprintf(stdout,"</tr>\n<TR>\n");
			}

		/* Set our current starting X position to be the "lastX" */
		lastx = slices[i].ax;

		/* Determine our slice width and height */
		slicew = slices[i].bx -slices[i].ax +1;
		sliceh = slices[i].by -slices[i].ay +1;

		
		/* If we have a "LINK", then print in the ahref='' stuff */
		if (slices[i].link[0] != '\0') {
			fprintf(stdout,"<TD colspan=\"%d\"><a href=\"%s\"><img src=\"%s\" border=\"0\" width=\"%d\" height=\"%d\"></a></td>\n",cols,slices[i].link,foutname,slicew,sliceh);
		}
		/* ELSE if we don't have a link, just put out the normal array of information */ 
		else {
		fprintf(stdout,"<TD colspan=\"%d\"><img src=\"%s\" border=\"0\" width=\"%d\" height=\"%d\"></td>\n",cols,foutname,slicew, sliceh);
		}
		/* ELSE .... well, this is where we will add the JavaScript stuff later... */

		/* Cut out the picture, and save the file... */
		
		
		/* --- old non-imgmagick cut(slices[i].ax, slices[i].ay, slices[i].bx, slices[i].by); */
		
		
		/* Setup the ImageInformation as required by ImageMagick */
		peice_box.width = slicew;
		peice_box.height = sliceh;
		peice_box.x = slices[i].ax;
		peice_box.y = slices[i].ay;

		/* Setup the info struct for our big image, so as to keep all the quality settings etc. */		
		peice_info = CloneImageInfo((ImageInfo *) NULL);		
		
		/* cut out our little peice */
		peice = CropImage( image, &peice_box, &exception );
		
		/* if we had a bugger up, then let the world know */
      if (peice == (Image *) image_info)
        MagickError(exception.severity,exception.message,exception.qualifier);

		/* put in the file name of the output file */
		strcpy(peice_info->filename, foutname);
		strcpy(peice->filename, foutname);

		/* Write the file */		
		status = WriteImage(peice_info, peice);
		
		if (status != 1) fprintf(stderr,"Could not write %s\n",foutname);

		/* clean up */		
		DestroyImageInfo(peice_info);
		DestroyImage(peice);

		ffr = 0;
	}

	fprintf(stdout,"\n<!-- total width = %d -->\n",tally);
	
	fprintf(stdout,"</tr></table>");
			


	return 0;
	}
	



	/*
	 * filetostdout
	 *
	 * prints a given file to stdout
	 *
	 */
int filetostdout( char *fname ){

	FILE *f = fopen(fname,"r");
	char achar;
	
	/* if the file opened okay... */
	if (f) {
		while ( (achar = fgetc(f))!= EOF) fprintf(stdout,"%c",achar);
		fclose(f);
		}
	else {
		fprintf(stderr,"Error: could not open file %s\n",fname);
		}
		
	return 0;
	}
	
			
	
/*
 * main
 *
 * The basic procedure is as follows...
 *
 * read in the input jpeg,
 * x1,y1 -> x2, y2 coord's 
 * output file name, and 
 * the quality factor
 *
 * read in the required file
 * copy the required "slice" into a new buffer
 * write the new buffer to the new file name
 *
 */
int main( int argc, char **argv ) {

	int aindex; /* argv index */
	char *infile=NULL, *outfile=NULL, *scriptfile=NULL, *headerfile=NULL, *footerfile=NULL, *suffix=NULL;

	/* ImageMagick stuff */
	ExceptionInfo exception;
	Image *image;
	ImageInfo *image_info;
		

	/* 
	 * Get the command line parameters ... 
	 */
	for ( aindex = 1; aindex < argc; aindex++ ){
	
		if (argv[aindex][0] == '-'){
		
			if ( argv[aindex][1] == 'i' ){
				infile = argv[aindex+1];
				}
			else
			if ( argv[aindex][1] == 'p' ){
				outfile = argv[aindex+1];
				}
			else
			if ( argv[aindex][1] == 's' ){
				suffix = argv[aindex+1];
				}
			else
			if ( argv[aindex][1] == 'c' ){
				scriptfile = argv[aindex+1];
				}
			else			
			if ( argv[aindex][1] == 'q' ){
				/*quality = atoi(argv[aindex+1]);*/
				}
			else
			if ( argv[aindex][1] == 'h' ){
				headerfile = argv[aindex+1];
				}
			else
			if ( argv[aindex][1] == 'f' ){
				footerfile = argv[aindex+1];
				}
			else 
			if (strstr(argv[aindex],"help")){
				fprintf(stderr,_HELP);
				exit(1);
				}
			else {
			/* if nothing fitted... then dump HELP!! */
				fprintf(stderr,_HELP);
				exit(1);
				}
			} /* if this was a '-' argument */
		
		} /* for */
		
		
	/* 
	 * Check for errors in the input... 
	 *
	 */
	
	if (!infile) {
		fprintf(stderr,"Error: I -need- an input JPEG file, use the -i option\n");
		exit(1);
		}
		
	if (!outfile) {
		fprintf(stderr,"Error: output file PREFIX required, use the -p option\n");
		exit(1);
		}
		
	if (!scriptfile) {
		fprintf(stderr,"Error: cutlist file not specified, use the -c option\n");		
		exit(1);
		}
		
/*
	if (!quality) {
		fprintf(stderr,"Warning: quality not specified, using %d%%\n",_QUALITY_DEFAULT);
		quality = _QUALITY_DEFAULT;
		}
			
	if ((quality > 100)||(quality < 1)) {
		fprintf(stderr,"Invalid JPEG quality setting, please choose a value between 1 and 100\n");
		exit(1);
		}
*/
	/* if the user didn't supply us with a suffix, then we'll have
	 * to take if off the input file 
	 */

#ifdef DEBUG
fprintf(stderr,"Testing suffix...\n");
#endif
	if (!suffix) {
		suffix = strrchr( infile, '.');
		}	 

	if (!suffix) suffix = _default_suffix;

	/* Lets try get this image file and read it in etc */
	MagickIncarnate(*argv);
   GetExceptionInfo(&exception);
   image_info=CloneImageInfo((ImageInfo *) NULL);
   (void) strcpy(image_info->filename, infile);
   image=ReadImage(image_info,&exception);
   if (image == (Image *) NULL)
   	MagickError(exception.severity,exception.message,exception.qualifier);	 
 
	/* Test to see if this picture isn't too large for our jslice program
	 */
	image_width = image->columns;
	image_height = image->rows;
	if ((image_width >= MAX_WIDTH) || (image_height >= MAX_HEIGHT)){
		fprintf(stderr,"Sorry, but jslice has been compiled only to handle pictures up to %d x %d pixels\n",MAX_WIDTH,MAX_HEIGHT);
		exit(1);
		}
	
	/*
	 * We've gotten this far... now lets actually produce some results
	 */

	/* Print my HTML " I did this " message*/
	fprintf(stdout,_MESSAGE);		

	/* add in the potential header HTML (because the default jslice output is not
	 * a complete HTML file */
	if (headerfile) {
		filetostdout(headerfile);
		}

	/* Slice up and dump out our jpeg pics and our HTML table */
	script_cut(image, image_info, scriptfile, outfile, suffix );
	
	/* add in the potential footer HTML */
	if (footerfile) {
		filetostdout(footerfile);
		}

	/* clean up the main image details... */
	DestroyImageInfo(image_info);
	DestroyImage(image);
			
	return 0;
	}
